%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Efficient ERB-MDCT analysis and synthesis                    %
% and plotting time-frequency image                            %
% Olivier Derrien - LMA/CNRS - 02.2015                         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear all;
close all;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Read the audio signal
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

name = 'svega.wav';
[x,Fs] = wavread(name);
x = x';
Lx = length(x);
fprintf('Signal length is: %d\n',Lx)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute ERB_MDCT parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Basic constants
v = 3; % number of bands per ERB (positive integer)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Generation of the ERB scale
dE = 1/v;
Imax = round(v*9.265*log(1+Fs/2/228.8455));
Emax = Imax/v;
alpha = Fs/2/(exp(Emax/9.265)-1);
E = dE:dE:Emax;
f_center = alpha*(exp(E/9.265)-1);
k_center = 2*Lx*f_center/Fs;

% Computing the bandwidths (in MDCT index)
k = [k_center(1), k_center];
K = length(k);
Delta = ones(K,K);
Delta = tril(Delta);
Delta(2:end,1) = 0.5;
temp = diag(0.5*ones(1,K));
Delta = Delta-temp;
Delta(1,1) = 1;

% Loop to get Nt as close a possible to Lx
vL = (Lx-50) : 0.1 : (Lx+50);
Nt = zeros(size(vL));
for p = 1 : length(vL)
    Nr = inv(Delta)*k';
    N = round(Nr/2)*2;
    Nt(p) = Delta(end,:)*N;
end
[val,idx] = min(abs(Nt-Lx));
Nr = inv(Delta)*k';
N = round(Nr/2)*2;
Nt = Delta(end,:)*N;

% Zero padding/truncating to get exactly Lx = Nt
if Nt <= Lx
    x = x(1:Nt);
else
    x = [x zeros(1,Nt-Lx)];
end
Lx = length(x);
N = N';

fprintf('The effective length of atomes is %d\n',Nt)
fprintf('Total number of transform coefficients is %d\n',sum(N))
fprintf('Redundancy is %e\n',sum(N)/Nt)

% Compute intermediate constants for variable-size MDCT
Ng = [N(1) N(1:end-1)]; % length of sine archs, left part
Nd = N;  % length of sine archs, right part
N1 = [0 (N(2:end)-N(1:end-1))/2]; % number of ones in the middle
ofs=zeros(1,K); % offset of the windows in the frequency domain
for p = 2:K
   ofs(p) = ofs(p-1)+N(p-1)-N1(p); 
end
Nt_pad = ofs(K)+2*N1(K)+N(K-1)+N(K); % total length of atoms with zero padding
zg = ofs+N1; % length of zero padding, left part
zd = Nt_pad-(zg+Ng+Nd+N1); % length of zero padding, right part

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Analysis part
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Apply the DCT-IV to the whole signal
NX = Lx+N(1)+N(end);
X = zeros(1,NX);
X(N(1)+1:N(1)+Lx) = DCTIV(x);
X(1:N(1)) = fliplr(X(N(1)+1:2*N(1)));
X(N(1)+Lx+1:end) = -fliplr(X(N(1)+Lx-N(end)+1:N(1)+Lx));

% Apply variable-size MDCTs and computation of TF image
X_ERB_MDCT = cell(1,K); % ERB-MDCT coefficients
sub_visu = 64; % sub-sampling for visualization
Nt_visu = ceil(Nt/sub_visu); % size of plotted TF image
TF_visu = zeros(K,Nt_visu); % plotted TF image
for p = 1 : K
    
    % Generating the MDCT window
    w_NS_MDCT = [zeros(1,N1(p)), sin(pi/2/Ng(p)*([0:Ng(p)-1]+0.5)), ones(1,N1(p)), cos(pi/2/Nd(p)*([0:Nd(p)-1]+0.5))];
        
    % Extracting a part of DCT-IV spectrum
    X_win = X(ofs(p)+1:ofs(p)+2*N(p));
    
    % Applying MDCT
    coeffs = MDCT(X_win.*w_NS_MDCT);
    X_ERB_MDCT{p} = coeffs;
        
    % Computing the interpolated TF image
    TF_visu(p,:) = interpft(20*log10(abs(coeffs)),Nt_visu);
        
end

% Plotting TF image
dB_min = -70; % lower threshold in dB
TF_visu = TF_visu - max(max(TF_visu));
TF_visu = max(TF_visu, dB_min);
[Ymax,Xmax] = size(TF_visu);
xscale = (0:Xmax-1)/Xmax*Nt_visu/Fs*sub_visu;
yscale = (0:Ymax-1)/Ymax*Fs/2000;
figure(1)
imagesc(xscale,yscale,TF_visu)
set(gca,'YDir','normal')
colorbar
xlabel('Time (s)')
ylabel('Frequency (kHz)')
title('ERB-MDCT energy (dB)')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Synthesis part
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Applying variable-size inverse MDCTs
X_rec = zeros(size(X));
for p = 1 : K
    
    % Generating the MDCT window
    w_NS_MDCT = [zeros(1,N1(p)), sin(pi/2/Ng(p)*([0:Ng(p)-1]+0.5)), ones(1,N1(p)), cos(pi/2/Nd(p)*([0:Nd(p)-1]+0.5))];
    
    % Applying inverse MDCT
    coeffs = X_ERB_MDCT{p};
    X_rec_win = iMDCT(coeffs).*w_NS_MDCT;
        
    % Reconstructing a part of DCT-IV spectrum
    X_rec(ofs(p)+1:ofs(p)+2*N(p)) = X_rec(ofs(p)+1:ofs(p)+2*N(p)) + X_rec_win;
        
end

% Trucating the DCT-IV spectrum and applying inverse DCT-IV
X_rec = X_rec(N(1)+1:N(1)+Lx);
x_rec = DCTIV(X_rec);

% Plotting the reconstructed signal and the reconstruction error
figure(2)
time = (0:Lx-1)/Fs;
plot(time,x), hold on
plot(time,x_rec,'r')
plot(time,x-x_rec,'g'), hold off
xlabel('Time (s)')
ylabel('Aplitude')
legend('Original signal','Reconstructed signal','Reconstruction error')

% Computing the SER in dB
SER_dB = 10*log10(var(x)/var(x-x_rec));
fprintf('The signal-to-error ratio is %d dB\n',round(SER_dB));

